"""Run this via `make climate_categories/data/CRF1999.yaml` in the main
directory."""

import pathlib

import climate_categories

OUTPATH = pathlib.Path("./climate_categories/data/CRF1999.yaml")


def main():
    """Create the CRF1999 categorization from the IPCC1996 categorization, which was
    also used by the UNFCCC to develop CRF1999. However, we can't use extend() and
    CRF1999 is not considered an extension of IPCC1996, because some codes were
    re-used for other categories.

    The CRF1999 categorization contains categories without codes. For new sets of
    categories, I formed codes as usual. However, some categories are inserted at parts
    of the hierarchy where this is not possible. For those, I
    generate codes from the parent codes plus a dash followed by the first three letters
    of the category title. Better than nothing.
    """
    spec = climate_categories.IPCC1996.to_spec()

    # Metadata
    spec["name"] = "CRF1999"
    spec["title"] = "Common Reporting Format GHG emissions categories (1999)"
    spec["comment"] = (
        "Classification of green-house gas emissions and removals into"
        " categories for use in annual inventories using the Common"
        " Reporting Format as specified in the UNFCCC guidelines on"
        " reporting"
        " and review as decided in the fifth session of the Conference of"
        " the Parties in 1999"
    )
    spec["references"] = (
        "United Nations 2000, Review of the Implementation of"
        " Commitments and of Other Provisions of the Convention,"
        " Annex Common Reporting Format, UNDOC FCCC/CP/1999/7, pages"
        " 15ff, https://undocs.org/en/FCCC/CP/1999/7"
    )
    spec["institution"] = "UN"
    spec["last_update"] = "1999-11-05"
    spec["version"] = "1999"
    spec["total_sum"] = True
    spec["canonical_top_level_category"] = "0"

    # Changes in categories
    cats = spec["categories"]
    ncats = {}

    # page 32 - reference approach is an alternative to sectoral approach
    ncats["1.A-ref"] = {
        "title": "Reference Approach",
        "children": [["1.A-ref.1", "1.A-ref.2"]],
    }
    cats["1.A"]["children"].append(["1.A-ref"])

    ncats["1.A-ref.1"] = {
        "title": "Fossil Fuel",
        "children": [["1.A-ref.1.a", "1.A-ref.1.b", "1.A-ref.1.c"]],
    }
    ncats["1.A-ref.1.a"] = {
        "title": "Liquid Fossil",
        "children": [["1.A-ref.1.a.i", "1.A-ref.1.a.ii"]],
    }

    ncats["1.A-ref.1.a.i"] = {
        "title": "Primary Fuels",
        "children": [["1.A-ref.1.a.i.1", "1.A-ref.1.a.i.2", "1.A-ref.1.a.i.3"]],
    }
    ncats["1.A-ref.1.a.i.1"] = {"title": "Crude Oil"}
    ncats["1.A-ref.1.a.i.2"] = {"title": "Orimulsion"}
    ncats["1.A-ref.1.a.i.3"] = {"title": "Natural Gas Liquids"}

    ncats["1.A-ref.1.a.ii"] = {
        "title": "Secondary Fuels",
        "children": [[f"1.A-ref.1.a.ii.{x}" for x in range(1, 15)]],
    }
    ncats["1.A-ref.1.a.ii.1"] = {"title": "Gasoline"}
    ncats["1.A-ref.1.a.ii.2"] = {"title": "Jet Kerosene"}
    ncats["1.A-ref.1.a.ii.3"] = {"title": "Other Kerosene"}
    ncats["1.A-ref.1.a.ii.4"] = {"title": "Shale Oil"}
    ncats["1.A-ref.1.a.ii.5"] = {"title": "Gas / Diesel Oil"}
    ncats["1.A-ref.1.a.ii.6"] = {"title": "Residual Fuel Oil"}
    ncats["1.A-ref.1.a.ii.7"] = {"title": "LPG"}
    ncats["1.A-ref.1.a.ii.8"] = {"title": "Ethane"}
    ncats["1.A-ref.1.a.ii.9"] = {"title": "Naphtha"}
    ncats["1.A-ref.1.a.ii.10"] = {"title": "Bitumen"}
    ncats["1.A-ref.1.a.ii.11"] = {"title": "Lubricants"}
    ncats["1.A-ref.1.a.ii.12"] = {"title": "Petroleum Coke"}
    ncats["1.A-ref.1.a.ii.13"] = {"title": "Refinery Feedstocks"}
    ncats["1.A-ref.1.a.ii.14"] = {"title": "Other Oil"}

    ncats["1.A-ref.1.b"] = {
        "title": "Solid Fossil",
        "children": [["1.A-ref.1.b.i", "1.A-ref.1.b.ii"]],
    }

    ncats["1.A-ref.1.b.i"] = {
        "title": "Primary Fuels",
        "children": [[f"1.A-ref.1.b.i.{x}" for x in range(1, 8)]],
    }
    ncats["1.A-ref.1.b.i.1"] = {"title": "Anthracite"}
    ncats["1.A-ref.1.b.i.2"] = {"title": "Coking Coal"}
    ncats["1.A-ref.1.b.i.3"] = {"title": "Other Bituminous Coal"}
    ncats["1.A-ref.1.b.i.4"] = {"title": "Sub-bituminous coal"}
    ncats["1.A-ref.1.b.i.5"] = {"title": "Lignite"}
    ncats["1.A-ref.1.b.i.6"] = {"title": "Oil Shale"}
    ncats["1.A-ref.1.b.i.7"] = {"title": "Peat"}

    ncats["1.A-ref.1.b.ii"] = {
        "title": "Secondary Fuels",
        "children": [["1.A-ref.1.b.ii.1", "1.A-ref.1.b.ii.2"]],
    }
    ncats["1.A-ref.1.b.ii.1"] = {"title": "BKB & Patent Fuel"}
    ncats["1.A-ref.1.b.ii.2"] = {"title": "Coke Oven / Gas Coke"}

    ncats["1.A-ref.1.c"] = {"title": "Gaseous Fossil", "children": [["1.A-ref.1.c.i"]]}
    ncats["1.A-ref.1.c.i"] = {"title": "Natural Gas (Dry)"}

    ncats["1.A-ref.2"] = {
        "title": "Biomass",
        "children": [["1.A-ref.2.a", "1.A-ref.2.b", "1.A-ref.2.c"]],
    }
    ncats["1.A-ref.2.a"] = {"title": "Solid Biomass"}
    ncats["1.A-ref.2.b"] = {"title": "Liquid Biomass"}
    ncats["1.A-ref.2.c"] = {"title": "Gas Biomass"}

    # page 36
    ncats["1.B.2.b-exp"] = {"title": "Exploration"}
    ncats["1.B.2.b-dis"] = {"title": "Distribution"}
    cats["1.B.2.b"]["children"] = [
        ["1.B.2.b-exp", "1.B.2.b.i", "1.B.2.b.ii", "1.B.2.b-dis", "1.B.2.b.iii"]
    ]

    ncats["1.B.2.b.iii.1"] = {"title": "at industrial plants and power stations"}
    ncats["1.B.2.b.iii.2"] = {"title": "in residential and commercial sectors"}
    cats["1.B.2.b.iii"]["children"] = [["1.B.2.b.iii.1", "1.B.2.b.iii.2"]]

    ncats["1.B.2.c-ven"] = {
        "title": "Venting",
        "children": [["1.B.2.c-ven.i", "1.B.2.c-ven.ii", "1.B.2.c-ven.iii"]],
    }
    ncats["1.B.2.c-ven.i"] = {"title": "Oil"}
    ncats["1.B.2.c-ven.ii"] = {"title": "Gas"}
    ncats["1.B.2.c-ven.iii"] = {"title": "Combined"}
    ncats["1.B.2.c-fla"] = {
        "title": "Flaring",
        "children": [["1.B.2.c-fla.i", "1.B.2.c-fla.ii", "1.B.2.c-fla.iii"]],
    }
    ncats["1.B.2.c-fla.i"] = {"title": "Oil"}
    ncats["1.B.2.c-fla.ii"] = {"title": "Gas"}
    ncats["1.B.2.c-fla.iii"] = {"title": "Combined"}
    cats["1.B.2.c"]["children"] = [["1.B.2.c-ven", "1.B.2.c-fla"]]
    for cat in ("1.B.2.c.i", "1.B.2.c.ii", "1.B.2.c.iii"):
        del cats[cat]

    ncats["1.B.2.d"] = {"title": "Other"}
    cats["1.B.2"]["children"][0].append("1.B.2.d")

    # page 39
    ncats["2.E.1.a"] = {"title": "Production of HCFC-22"}
    ncats["2.E.1.b"] = {"title": "Other"}
    cats["2.E.1"]["children"] = [["2.E.1.a", "2.E.1.b"]]

    # re-numbered
    cats["2.F.8"] = cats["2.F.6"]
    del cats["2.F.6"]
    cats["2.F.8"]["alternative_codes"] = ["2F8", "2 F 8"]

    cats["2.F.4"]["title"] = "Aerosols / Metered Dose Inhalers"
    ncats["2.F.6"] = {"title": "Semiconductor Manufacture"}
    ncats["2.F.7"] = {"title": "Electrical Equipment"}
    cats["2.F"]["children"] = [[f"2.F.{x}" for x in range(1, 9)]]

    # page 40
    ncats["2.A.7.a"] = {"title": "Glass Production"}
    ncats["2.A.7.b"] = {"title": "Other"}
    cats["2.A.7"]["children"] = [["2.A.7.a", "2.A.7.b"]]

    ncats["2.B.4.a"] = {"title": "Silicon Carbide"}
    ncats["2.B.4.b"] = {"title": "Calcium Carbide"}
    cats["2.B.4"]["children"] = [["2.B.4.a", "2.B.4.b"]]

    ncats["2.B.5.a"] = {"title": "Carbon Black"}
    ncats["2.B.5.b"] = {"title": "Ethylene"}
    ncats["2.B.5.c"] = {"title": "Dichloroethylene"}
    ncats["2.B.5.d"] = {"title": "Styrene"}
    ncats["2.B.5.e"] = {"title": "Methanol"}
    ncats["2.B.5.f"] = {"title": "Other"}
    cats["2.B.5"]["children"] = [[f"2.B.5.{x}" for x in "abcdef"]]

    # page 41
    ncats["2.C.1.a"] = {"title": "Steel"}
    ncats["2.C.1.b"] = {"title": "Pig Iron"}
    ncats["2.C.1.c"] = {"title": "Sinter"}
    ncats["2.C.1.d"] = {"title": "Coke"}
    ncats["2.C.1.e"] = {"title": "Other"}
    cats["2.C.1"]["children"] = [[f"2.C.1.{x}" for x in "abcde"]]

    # page 42
    ncats["2.C.4.a"] = {"title": "SF6 used in Aluminium Foundries"}
    ncats["2.C.4.b"] = {"title": "SF6 used in Magnesium Foundries"}
    cats["2.C.4"]["children"] = [["2.C.4.a", "2.C.4.b"]]

    # Page 45
    ncats["2.F.1.a"] = {"title": "Domestic Refrigeration"}
    ncats["2.F.1.b"] = {"title": "Commercial Refrigeration"}
    ncats["2.F.1.c"] = {"title": "Transport Refrigeration"}
    ncats["2.F.1.d"] = {"title": "Industrial Refrigeration"}
    ncats["2.F.1.e"] = {"title": "Stationary Air-Conditioning"}
    ncats["2.F.1.f"] = {"title": "Mobile Air-Conditioning"}
    cats["2.F.1"]["children"] = [[f"2.F.1.{x}" for x in "abcdef"]]

    ncats["2.F.2.a"] = {"title": "Hard Foam"}
    ncats["2.F.2.b"] = {"title": "Soft Foam"}
    cats["2.F.2"]["children"] = [["2.F.2.a", "2.F.2.b"]]

    # Page 46
    ncats["2.F.4.a"] = {"title": "Metered Dose Inhalers"}
    ncats["2.F.4.b"] = {"title": "Other"}
    cats["2.F.4"]["children"] = [["2.F.4.a", "2.F.4.b"]]

    # Page 47
    ncats["3.D.1"] = {"title": "Use of N2O for Anaesthesia"}
    ncats["3.D.2"] = {"title": "N2O from Fire Extinguishers"}
    ncats["3.D.3"] = {"title": "N2O from Aerosol Cans"}
    ncats["3.D.4"] = {"title": "Other Use of N2O"}
    ncats["3.D.5"] = {"title": "Other"}
    cats["3.D"]["children"] = [[f"3.D.{x}" for x in range(1, 6)]]

    # Page 55
    ncats["4.D.1"] = {
        "title": "Direct Soil Emissions",
        "children": [[f"4.D.1.{x}" for x in "abcde"]],
    }
    ncats["4.D.1.a"] = {"title": "Synthetic Fertilizers"}
    ncats["4.D.1.b"] = {"title": "Animal Wastes Applied to Soils"}
    ncats["4.D.1.c"] = {"title": "N-fixing Crops"}
    ncats["4.D.1.d"] = {"title": "Crop Residue"}
    ncats["4.D.1.e"] = {"title": "Cultivation of Histosols"}

    ncats["4.D.2"] = {"title": "Animal Production"}

    ncats["4.D.3"] = {
        "title": "Indirect Emissions",
        "children": [["4.D.3.a", "4.D.3.b"]],
    }
    ncats["4.D.3.a"] = {"title": "Atmospheric Deposition"}
    ncats["4.D.3.b"] = {"title": "Nitrogen Leaching and Run-off"}
    ncats["4.D.4"] = {"title": "Other"}

    cats["4.D"]["children"] = [["4.D.1", "4.D.2", "4.D.3", "4.D.4"]]

    # Page 57
    ncats["4.F.1.a"] = {"title": "Wheat"}
    ncats["4.F.1.b"] = {"title": "Barley"}
    ncats["4.F.1.c"] = {"title": "Maize"}
    ncats["4.F.1.d"] = {"title": "Oats"}
    ncats["4.F.1.e"] = {"title": "Rye"}
    ncats["4.F.1.f"] = {"title": "Rice"}
    ncats["4.F.1.g"] = {"title": "Other"}
    cats["4.F.1"]["children"] = [[f"4.F.1.{x}" for x in "abcdefg"]]

    ncats["4.F.2.a"] = {"title": "Dry bean"}
    ncats["4.F.2.b"] = {"title": "Peas"}
    ncats["4.F.2.c"] = {"title": "Soybeans"}
    ncats["4.F.2.d"] = {"title": "Other"}
    cats["4.F.2"]["children"] = [[f"4.F.2.{x}" for x in "abcd"]]

    ncats["4.F.3.a"] = {"title": "Potatoes"}
    ncats["4.F.3.b"] = {"title": "Other"}
    cats["4.F.3"]["children"] = [["4.F.3.a", "4.F.3.b"]]

    # Page 58ff
    # changed a lot, re-do completely
    for cat in (
        "5.A",
        "5.A.1",
        "5.A.1.a",
        "5.A.1.b",
        "5.A.1.c",
        "5.A.1.d",
        "5.A.1.e",
        "5.A.1.f",
        "5.A.1.g",
        "5.A.1.h",
        "5.A.2",
        "5.A.2.a",
        "5.A.2.b",
        "5.A.2.c",
        "5.A.2.d",
        "5.A.3",
        "5.A.3.a",
        "5.A.3.b",
        "5.A.3.c",
        "5.A.4",
        "5.A.5",
    ):
        del cats[cat]

    ncats["5.A"] = {
        "title": "Changes in Forest and Other Woody Biomass Stocks",
        "children": [["5.A.1", "5.A.2", "5.A.3", "5.A.4", "5.A.5"]],
    }
    ncats["5.A.1"] = {
        "title": "Tropical",
        "children": [["5.A.1.a", "5.A.1.b", "5.A.1.c"]],
    }

    ncats["5.A.1.a"] = {
        "title": "Plantations",
        "children": [
            [
                f"5.A.1.a.{x}"
                for x in ("i", "ii", "iii", "iv", "v", "vi", "vii", "viii", "ix")
            ]
        ],
    }
    ncats["5.A.1.a.i"] = {"title": "Acacia spp."}
    ncats["5.A.1.a.ii"] = {"title": "Eucalyptus spp."}
    ncats["5.A.1.a.iii"] = {"title": "Tectona grandis"}
    ncats["5.A.1.a.iv"] = {"title": "Pinus spp"}
    ncats["5.A.1.a.v"] = {"title": "Pinus caribaea"}
    ncats["5.A.1.a.vi"] = {"title": "Mixed Hardwoods"}
    ncats["5.A.1.a.vii"] = {"title": "Mixed Fast-Growing Hardwoods"}
    ncats["5.A.1.a.viii"] = {"title": "Mixed Softwoods"}
    ncats["5.A.1.a.ix"] = {"title": "Other"}

    ncats["5.A.1.b"] = {
        "title": "Other Forests",
        "children": [["5.A.1.b.i", "5.A.1.b.ii", "5.A.1.b.iii"]],
    }
    ncats["5.A.1.b.i"] = {"title": "Moist"}
    ncats["5.A.1.b.ii"] = {"title": "Seasonal"}
    ncats["5.A.1.b.iii"] = {"title": "Dry"}

    ncats["5.A.1.c"] = {"title": "Other"}

    ncats["5.A.2"] = {
        "title": "Temperate",
        "children": [["5.A.2.a", "5.A.2.b", "5.A.2.c"]],
    }
    ncats["5.A.2.a"] = {"title": "Plantations"}

    ncats["5.A.2.b"] = {
        "title": "Commercial",
        "children": [["5.A.2.b.i", "5.A.2.b.ii"]],
    }
    ncats["5.A.2.b.i"] = {"title": "Evergreen"}
    ncats["5.A.2.b.ii"] = {"title": "Deciduous"}

    ncats["5.A.2.c"] = {"title": "Other"}

    ncats["5.A.3"] = {"title": "Boreal"}
    ncats["5.A.4"] = {"title": "Grasslands / Tundra"}

    ncats["5.A.5"] = {"title": "Other", "children": [["5.A.5.a", "5.A.5.b"]]}
    ncats["5.A.5.a"] = {"title": "Harvested Wood"}
    ncats["5.A.5.b"] = {"title": "Other"}

    # Page 58
    ncats["5.D.1"] = {"title": "Cultivation of Mineral Soils"}
    ncats["5.D.2"] = {"title": "Cultivation of Organic Soils"}
    ncats["5.D.3"] = {"title": "Liming of Agricultural Soils"}
    ncats["5.D.4"] = {"title": "Forest Soils"}
    ncats["5.D.5"] = {"title": "Other"}
    cats["5.D"]["children"] = [[f"5.D.{x}" for x in range(1, 6)]]

    # Page 60
    ncats["5.B-tro"] = {"title": "Tropical Savanna / Grasslands"}
    ncats["5.B-gra"] = {"title": "Grasslands"}
    cats["5.B"]["children"][0].append("5.B-tro")
    cats["5.B"]["children"][0].append("5.B-gra")

    ncats["5.B.2-mix"] = {"title": "Mixed Broadleaf / Coniferous"}
    cats["5.B.2"]["children"][0].append("5.B.2-mix")

    # Page 61
    ncats["5.C-tro"] = {"title": "Tropical Savanna / Grasslands"}
    ncats["5.C-gra"] = {"title": "Grasslands"}
    cats["5.C"]["children"][0].append("5.C-tro")
    cats["5.C"]["children"][0].append("5.C-gra")

    ncats["5.C.1.a"] = {"title": "Wet / Very Moist"}
    ncats["5.C.1.b"] = {"title": "Moist, short dry season"}
    ncats["5.C.1.c"] = {"title": "Moist, long dry season"}
    ncats["5.C.1.d"] = {"title": "Dry"}
    ncats["5.C.1.e"] = {"title": "Montane Moist"}
    ncats["5.C.1.f"] = {"title": "Montane Dry"}
    cats["5.C.1"]["children"] = [[f"5.C.1.{x}" for x in "abcdef"]]

    for parent in ("5.C.2", "5.C.3"):
        ncats[f"{parent}.a"] = {"title": "Mixed Broadleaf / Coniferous"}
        ncats[f"{parent}.b"] = {"title": "Coniferous"}
        ncats[f"{parent}.c"] = {"title": "Broadleaf"}
        cats[parent]["children"] = [[f"{parent}.{x}" for x in "abc"]]

    # Page 62
    ncats["5.D.1"]["children"] = [[f"5.D.1.{x}" for x in "abcdef"]]
    ncats["5.D.1.a"] = {"title": "High Activity Soils"}
    ncats["5.D.1.b"] = {"title": "Low Activity Soils"}
    ncats["5.D.1.c"] = {"title": "Sandy"}
    ncats["5.D.1.d"] = {"title": "Volcanic"}
    ncats["5.D.1.e"] = {"title": "Wetland (Aquic)"}
    ncats["5.D.1.f"] = {"title": "Other"}

    ncats["5.D.2"]["children"] = [["5.D.2.a", "5.D.2.b", "5.D.2.c"]]
    ncats["5.D.2.a"] = {
        "title": "Cool Temperate",
        "children": [["5.D.2.a.i", "5.D.2.a.ii"]],
    }
    ncats["5.D.2.a.i"] = {"title": "Upland Crops"}
    ncats["5.D.2.a.ii"] = {"title": "Pasture/Forest"}
    ncats["5.D.2.b"] = {
        "title": "Warm Temperate",
        "children": [["5.D.2.b.i", "5.D.2.b.ii"]],
    }
    ncats["5.D.2.b.i"] = {"title": "Upland Crops"}
    ncats["5.D.2.b.ii"] = {"title": "Pasture/Forest"}
    ncats["5.D.2.c"] = {"title": "Tropical", "children": [["5.D.2.c.i", "5.D.2.c.ii"]]}
    ncats["5.D.2.c.i"] = {"title": "Upland Crops"}
    ncats["5.D.2.c.ii"] = {"title": "Pasture/Forest"}

    ncats["5.D.3"]["children"] = [["5.D.3.a", "5.D.3.b"]]
    ncats["5.D.3.a"] = {"title": "Limestone"}
    ncats["5.D.3.b"] = {"title": "Dolomite"}

    # Page 64
    ncats["6.A.2.a"] = {"title": "deep (>5 m)"}
    ncats["6.A.2.b"] = {"title": "shallow (<5 m)"}
    cats["6.A.2"]["children"] = [["6.A.2.a", "6.A.2.b"]]

    ncats["6.C.1"] = {"title": "biogenic"}
    ncats["6.C.2"] = {"title": "plastics"}
    ncats["6.C.3"] = {"title": "other"}
    cats["6.C"]["children"] = [["6.C.1", "6.C.2", "6.C.3"]]

    # Pages 66ff
    ncats["0"] = {
        "title": "Total National Emissions and Removals",
        "children": [["1", "2", "3", "4", "5", "6", "7"]],
    }
    ncats["M.Memo"] = {
        "title": "Memo Items",
        "children": [["M.Memo.Int", "M.Memo.Mult", "M.Memo.Bio"]],
    }
    ncats["M.Memo.Int"] = {
        "title": "International Bunkers",
        "children": [["M.Memo.Int.Avi", "M.Memo.Int.Mar"]],
    }
    ncats["M.Memo.Int.Avi"] = {"title": "Aviation"}
    ncats["M.Memo.Int.Mar"] = {"title": "Marine"}
    ncats["M.Memo.Mult"] = {"title": "Multilateral Operations"}
    ncats["M.Memo.Bio"] = {"title": "CO2 Emissions from Biomass"}

    for ncode in ncats:
        if "." in ncode:
            ncats[ncode]["alternative_codes"] = [
                ncode.replace(".", " "),
                ncode.replace(".", ""),
            ]

    cats.update(ncats)

    CRF1999 = climate_categories.HierarchicalCategorization.from_spec(spec)

    CRF1999.to_yaml(OUTPATH)

    climate_categories.HierarchicalCategorization.from_yaml(OUTPATH)


if __name__ == "__main__":
    main()



"""Tests for the 'gas' categorization."""

import climate_categories


def test_categories():
    assert climate_categories.gas["CO2"].title == "carbon dioxide"
    assert (
        climate_categories.gas["HCFC414a"].comment
        == "The refrigerant HCFC414a, which is a mixture of 51% HCFC22, 28.5% HCFC124,"
        " 4% HC600a, 16.5% HCFC142b."
    )


def test_mixture_hierarchy():
    assert (
        climate_categories.gas["HFC417a"]
        in climate_categories.gas["mixtures"].children[0]
    )



import pytest

import climate_categories
import climate_categories._conversions as conversions
import climate_categories.tests
import climate_categories.tests.data


@pytest.fixture
def simple_conversion_specs():
    cat_a = {
        "name": "A",
        "title": "Categorization A",
        "canonical_top_level_category": "0",
        "hierarchical": True,
        "last_update": "2021-07-29",
        "comment": "The first categorization",
        "references": "ask mika",
        "institution": "PIK",
        "total_sum": True,
        "categories": {
            "0": {"title": "total", "children": [["1", "2"]]},
            "1": {"title": "one", "children": [["1.A", "1.B"]]},
            "2": {"title": "two", "children": [["2.A", "2.B"]]},
            "1.A": {"title": "one-A"},
            "1.B": {"title": "one-B"},
            "2.A": {"title": "two-A", "children": [["2.A.1"]]},
            "2.B": {"title": "two-B"},
            "2.A.1": {"title": "two-A-one"},
        },
    }
    cat_b = {
        "name": "B",
        "title": "Categorization B",
        "canonical_top_level_category": "0",
        "hierarchical": True,
        "last_update": "2021-07-29",
        "comment": "The second categorization",
        "references": "ask mika",
        "institution": "PIK",
        "total_sum": True,
        "categories": {
            "0": {"title": "total", "children": [["1", "2"]]},
            "1": {"title": "one", "children": [["1.A", "1.B"]]},
            "2": {"title": "two", "children": [["2.A"]]},
            "1.A": {"title": "one-A"},
            "1.B": {"title": "one-B"},
            "2.A": {"title": "two-A", "children": [["2.A.1", "2.A.2"]]},
            "2.A.1": {"title": "two-A-one", "children": [["2.A.1.a"]]},
            "2.A.2": {"title": "two-A-two"},
            "2.A.1.a": {"title": "two-A-one-a"},
        },
    }
    convs = [
        ["0", "0", "total"],
        ["1", "1"],
        ["1.A", "1.A"],
        ["1.B", "1.B"],
        ["2", "2"],
        ["2.A + 2.B", "2.A"],
        ["2.A", "2.A.1"],
        ["2.A.1", "2.A.1.a"],
        ["2.B", "2.A.2"],
    ]
    return cat_a, cat_b, convs


def specs_to_conversion(cat_a_spec, cat_b_spec, convs_spec) -> conversions.Conversion:
    cat_a = climate_categories.HierarchicalCategorization.from_spec(cat_a_spec)
    cat_b = climate_categories.HierarchicalCategorization.from_spec(cat_b_spec)
    return conversions.Conversion(
        categorization_a=cat_a,
        categorization_b=cat_b,
        rules=[
            conversions.ConversionRuleSpec.from_csv_row(
                iter(row), aux_names=[]
            ).hydrate(
                categorization_a=cat_a,
                categorization_b=cat_b,
                cats=climate_categories.cats,
            )
            for row in convs_spec
        ],
    )


def test_no_over_counting(simple_conversion_specs):
    conv = specs_to_conversion(*simple_conversion_specs)
    assert not conv.find_over_counting_problems()
    assert not conv.reversed().find_over_counting_problems()


def test_simple_over_counting(simple_conversion_specs):
    cat_a, cat_b, convs = simple_conversion_specs
    convs.append(["2", "1"])
    conv = specs_to_conversion(cat_a, cat_b, convs)
    problems = conv.find_over_counting_problems()
    problematic_categories = [problem.category for problem in problems]
    assert conv.categorization_a["2"] in problematic_categories
    assert conv.categorization_a["2.A"] in problematic_categories
    assert conv.categorization_a["2.A.1"] in problematic_categories
    assert conv.categorization_a["2.B"] in problematic_categories

    assert conv.categorization_b["1"] in problematic_categories
    assert conv.categorization_b["1.A"] in problematic_categories
    assert conv.categorization_b["1.B"] in problematic_categories
    assert len(problems) == 7


def test_over_counting_aux(simple_conversion_specs):
    cat_a_spec, cat_b_spec, convs_spec = simple_conversion_specs
    convs_spec.append(["2", "1"])
    for c in convs_spec:
        c.insert(1, "")
    cat_a = climate_categories.HierarchicalCategorization.from_spec(cat_a_spec)
    cat_b = climate_categories.HierarchicalCategorization.from_spec(cat_b_spec)
    conv = conversions.Conversion(
        categorization_a=cat_a,
        categorization_b=cat_b,
        rules=[
            conversions.ConversionRuleSpec.from_csv_row(
                iter(row), aux_names=["gas"]
            ).hydrate(
                categorization_a=cat_a,
                categorization_b=cat_b,
                cats=climate_categories.cats,
            )
            for row in convs_spec
        ],
    )

    problems = conv.find_over_counting_problems()
    problematic_categories = [problem.category for problem in problems]
    assert conv.categorization_a["2"] in problematic_categories
    assert conv.categorization_a["2.A"] in problematic_categories
    assert conv.categorization_a["2.A.1"] in problematic_categories
    assert conv.categorization_a["2.B"] in problematic_categories

    assert conv.categorization_b["1"] in problematic_categories
    assert conv.categorization_b["1.A"] in problematic_categories
    assert conv.categorization_b["1.B"] in problematic_categories
    assert len(problems) == 7


def test_sum_over_counting(simple_conversion_specs):
    cat_a, cat_b, convs = simple_conversion_specs
    cat_b["categories"]["1.C"] = {"title": "one-C"}
    cat_b["categories"]["1"]["children"][0].append("1.C")
    convs.append(["1.A + 1.B", "1.C"])
    conv = specs_to_conversion(cat_a, cat_b, convs)
    problems = conv.find_over_counting_problems()
    problematic_categories = [problem.category for problem in problems]
    assert len(problems) == 2
    assert conv.categorization_a["1.A"] in problematic_categories
    assert conv.categorization_a["1.B"] in problematic_categories


def test_sum_no_over_counting(simple_conversion_specs):
    cat_a, cat_b, convs = simple_conversion_specs
    convs.append(["1.A + 1.B", "1"])
    conv = specs_to_conversion(cat_a, cat_b, convs)
    problems = conv.find_over_counting_problems()
    assert not problems


def test_not_hierarchical(simple_conversion_specs):
    cat_a_spec, cat_b_spec, convs_spec = simple_conversion_specs
    for cat in cat_a_spec["categories"].values():
        try:
            del cat["children"]
        except KeyError:
            pass
    for cat in cat_b_spec["categories"].values():
        try:
            del cat["children"]
        except KeyError:
            pass
    cat_a_spec["hierarchical"] = False
    cat_b_spec["hierarchical"] = False

    cat_a = climate_categories.Categorization.from_spec(cat_a_spec)
    cat_b = climate_categories.Categorization.from_spec(cat_b_spec)

    conv = conversions.Conversion(
        categorization_a=cat_a,
        categorization_b=cat_b,
        rules=[
            conversions.ConversionRuleSpec.from_csv_row(
                iter(row), aux_names=[]
            ).hydrate(
                categorization_a=cat_a,
                categorization_b=cat_b,
                cats=climate_categories.cats,
            )
            for row in convs_spec
        ],
    )

    with pytest.raises(ValueError, match="is not hierarchical"):
        conv.find_over_counting_problems()


def test_not_total_sum(simple_conversion_specs):
    cat_a_spec, cat_b_spec, convs_spec = simple_conversion_specs
    cat_a_spec["total_sum"] = False
    conv = specs_to_conversion(
        cat_a_spec=cat_a_spec, cat_b_spec=cat_b_spec, convs_spec=convs_spec
    )
    with pytest.raises(
        ValueError,
        match="it is not specified that the sum of a set of children equals the parent",
    ):
        conv.find_over_counting_problems()



"""Tests for `climate_categories` package."""

import datetime
import importlib
import importlib.resources
import pathlib

import pandas as pd
import pytest
import strictyaml

import climate_categories
import climate_categories.tests
import climate_categories.tests.data


class TestSimple:
    def test_meta(self, SimpleCat: climate_categories.Categorization):
        assert SimpleCat.name == "SimpleCat" == str(SimpleCat)
        assert SimpleCat.title == "Simple Categorization"
        assert SimpleCat.references == "doi:00000/00000"
        assert SimpleCat.last_update == datetime.date(2021, 2, 23)
        assert SimpleCat.institution == "PIK"
        assert SimpleCat.version == "1"
        assert (
            SimpleCat.comment
            == "A simple example categorization without relationships between"
            " categories"
        )
        assert SimpleCat.hierarchical is False
        assert (
            repr(SimpleCat)
            == "<Categorization SimpleCat 'Simple Categorization' with 4 categories>"
        )

    def test_categories(self, SimpleCat: climate_categories.Categorization):
        assert "1" in SimpleCat
        assert SimpleCat["1"].title == "Category 1"
        assert SimpleCat["1"].comment == "The first category"
        assert SimpleCat["1"].codes == ("1", "A", "CatA")
        assert SimpleCat["1"] == SimpleCat["A"] == SimpleCat["CatA"]
        assert SimpleCat["1"] != SimpleCat["2"]
        assert SimpleCat["unnumbered"].title == "The unnumbered category"
        assert repr(SimpleCat["1"]) == "<SimpleCat: '1'>"
        assert str(SimpleCat["1"]) == "1 Category 1"
        assert SimpleCat["1"].info == {
            "important_data": ["A", "B", "C"],
            "other_important_thing": "ABC",
        }
        assert SimpleCat["2"].info == {}

    def test_dict_like(self, SimpleCat: climate_categories.Categorization):
        assert "A" in SimpleCat
        assert list(SimpleCat.all_keys()) == [
            "1",
            "A",
            "CatA",
            "2",
            "B",
            "CatB",
            "3",
            "C",
            "CatC",
            "unnumbered",
        ]
        assert list(SimpleCat.keys()) == ["1", "2", "3", "unnumbered"]
        assert list(SimpleCat.items()) == [
            ("1", SimpleCat["1"]),
            ("2", SimpleCat["2"]),
            ("3", SimpleCat["3"]),
            ("unnumbered", SimpleCat["unnumbered"]),
        ]
        assert len(SimpleCat) == 4

    def test_comparisons(self, SimpleCat: climate_categories.Categorization):
        assert list(sorted(SimpleCat.values())) == [
            SimpleCat["1"],
            SimpleCat["2"],
            SimpleCat["3"],
            SimpleCat["unnumbered"],
        ]

        assert not SimpleCat["1"] == "1"
        assert not SimpleCat["1"] == SimpleCat["2"]
        assert SimpleCat["1"] == SimpleCat["1"]
        ext = SimpleCat.extend(
            name="ext",
            categories={"10": {"title": "ten"}},
            alternative_codes={"yksi": "1"},
        )
        assert SimpleCat["1"] == ext["1"]
        assert SimpleCat["2"] == ext["2"]
        assert ext["yksi"] < SimpleCat["2"]
        assert not ext["yksi"] < SimpleCat["1"]

        assert list(sorted(ext.values())) == [
            ext["1"],
            ext["2"],
            ext["3"],
            ext["10"],
            ext["unnumbered"],
        ]

    def test_df(self, SimpleCat: climate_categories.Categorization):
        expected = pd.DataFrame(
            index=["1", "2", "3", "unnumbered"],
            data={
                "title": [
                    "Category 1",
                    "Category 2",
                    "Category 3",
                    "The unnumbered category",
                ],
                "comment": [
                    "The first category",
                    "The second category",
                    "The third category",
                    None,
                ],
                "alternative_codes": [
                    ("A", "CatA"),
                    ("B", "CatB"),
                    ("C", "CatC"),
                    tuple(),
                ],
            },
        )
        pd.testing.assert_frame_equal(SimpleCat.df, expected)

    def test_extend(self, SimpleCat: climate_categories.Categorization):
        SimpleCat_ext = SimpleCat.extend(
            name="ext",
            categories={
                "4": {
                    "title": "Category 4",
                    "comment": "The fourth category",
                    "alternative_codes": ["D", "CatD"],
                },
                "t": {"title": "Category T"},
            },
            title=" title_ext",
            alternative_codes={"I": "1", "II": "2", "III": "3", "drai": "3", "IV": "4"},
        )
        assert SimpleCat_ext.name == "SimpleCat_ext"
        assert SimpleCat_ext.title == "Simple Categorization title_ext"
        assert SimpleCat_ext.references == ""
        assert SimpleCat_ext.last_update == datetime.date.today()
        assert SimpleCat_ext.institution == ""
        assert SimpleCat_ext.version == "1"
        assert (
            SimpleCat_ext.comment
            == "A simple example categorization without relationships between"
            " categories extended by ext"
        )
        assert SimpleCat_ext.hierarchical is False

        assert "1" in SimpleCat_ext
        assert SimpleCat_ext["1"].title == "Category 1"
        assert SimpleCat_ext["1"].comment == "The first category"
        assert SimpleCat_ext["1"].codes == ("1", "A", "CatA", "I")
        assert SimpleCat_ext["1"] == SimpleCat_ext["A"] == SimpleCat["CatA"]
        assert SimpleCat_ext["1"] != SimpleCat_ext["2"]
        assert SimpleCat_ext["unnumbered"].title == "The unnumbered category"
        assert repr(SimpleCat_ext["1"]) == "<SimpleCat_ext: '1'>"
        assert str(SimpleCat_ext["1"]) == "1 Category 1"

        assert "A" in SimpleCat_ext
        assert "D" in SimpleCat_ext
        assert SimpleCat_ext["1"] == SimpleCat_ext["I"]
        assert SimpleCat_ext["III"] == SimpleCat["3"] == SimpleCat_ext["drai"]
        assert list(SimpleCat_ext.all_keys()) == [
            "1",
            "A",
            "CatA",
            "I",
            "2",
            "B",
            "CatB",
            "II",
            "3",
            "C",
            "CatC",
            "III",
            "drai",
            "unnumbered",
            "4",
            "D",
            "CatD",
            "IV",
            "t",
        ]
        assert list(SimpleCat_ext.keys()) == ["1", "2", "3", "unnumbered", "4", "t"]
        assert len(SimpleCat_ext) == 6

    def test_extend_defaults(self, SimpleCat: climate_categories.Categorization):
        a = SimpleCat.extend(name="ext")
        assert a.name == "SimpleCat_ext"
        assert a.references == ""
        assert a.institution == ""
        assert a.title == "Simple Categorization + ext"
        assert (
            a.comment == "A simple example categorization without relationships between"
            " categories extended by ext"
        )
        assert a.last_update == datetime.date.today()

    def test_extend_not_defaults(self, SimpleCat: climate_categories.Categorization):
        b = SimpleCat.extend(
            name="ext",
            title="title",
            comment="comment",
            last_update=datetime.date.fromisoformat("2020-02-20"),
        )
        assert b.name == "SimpleCat_ext"
        assert b.references == ""
        assert b.institution == ""
        assert b.title == "Simple Categorizationtitle"
        assert (
            b.comment == "A simple example categorization without relationships between"
            " categoriescomment"
        )
        assert b.last_update == datetime.date.fromisoformat("2020-02-20")


class TestHierarchical:
    def test_meta(self, HierCat: climate_categories.HierarchicalCategorization):
        assert HierCat.name == "HierCat"
        assert HierCat.title == "Hierarchical Categorization"
        assert HierCat.references == "doi:00000/00000"
        assert HierCat.last_update == datetime.date(2021, 2, 23)
        assert HierCat.institution == "PIK"
        assert HierCat.version == "one"
        assert (
            HierCat.comment
            == "A simple hierarchical categorization with categories with relationships"
        )
        assert HierCat.hierarchical is True

    def test_categories(self, HierCat: climate_categories.HierarchicalCategorization):
        assert "0" in HierCat
        assert HierCat["0"].title == "Category 0"
        assert HierCat["0"].comment == "Top-most category"
        assert HierCat["0"].codes == ("0", "TOTAL")
        assert HierCat["0"] == HierCat["TOTAL"]
        assert HierCat["0"] != HierCat["1"]
        assert repr(HierCat["0"]) == "<HierCat: '0'>"
        assert str(HierCat["0"]) == "0 Category 0"
        assert HierCat["1"].info == {
            "SomeInfo": "A",
            "OtherInfo": ["A", "B", "C"],
        }
        assert HierCat["2"].info == {}

    def test_category_relationships(
        self, HierCat: climate_categories.HierarchicalCategorization
    ):
        assert HierCat["0"].children == [
            {HierCat["1"], HierCat["2"], HierCat["3"]},
            {HierCat["0X3"], HierCat["3"]},
            {HierCat["1A"], HierCat["1B"], HierCat["2"], HierCat["3"]},
        ]
        assert HierCat["1"].parents == {HierCat["0"], HierCat["0X3"]}
        assert HierCat.canonical_top_level_category == HierCat["0"]

        assert HierCat.level("0") == 1
        assert HierCat.level("1") == 2
        assert HierCat.level("2") == 2
        assert HierCat.level("0X3") == 2
        assert HierCat.level("1B") == 3

        assert HierCat.ancestors("1A") == {HierCat["1"], HierCat["0"], HierCat["0X3"]}
        assert HierCat.descendants("0X3") == {
            HierCat["1"],
            HierCat["2"],
            HierCat["1A"],
            HierCat["1B"],
            HierCat["2A"],
            HierCat["2B"],
        }

        with pytest.raises(
            ValueError,
            match="'OT' is not a transitive child of the canonical top level '0'.",
        ):
            HierCat.level("OT")

    def test_dict_like(self, HierCat: climate_categories.HierarchicalCategorization):
        assert "0" in HierCat
        assert list(HierCat.all_keys()) == [
            "0",
            "TOTAL",
            "1",
            "2",
            "3",
            "1A",
            "1a",
            "1B",
            "1b",
            "2A",
            "2a",
            "2B",
            "2b",
            "3A",
            "3a",
            "0X3",
            "0E3",
            "OT",
        ]
        assert list(HierCat.keys()) == [
            "0",
            "1",
            "2",
            "3",
            "1A",
            "1B",
            "2A",
            "2B",
            "3A",
            "0X3",
            "OT",
        ]
        assert list(HierCat.items()) == [
            ("0", HierCat["0"]),
            ("1", HierCat["1"]),
            ("2", HierCat["2"]),
            ("3", HierCat["3"]),
            ("1A", HierCat["1A"]),
            ("1B", HierCat["1B"]),
            ("2A", HierCat["2A"]),
            ("2B", HierCat["2B"]),
            ("3A", HierCat["3A"]),
            ("0X3", HierCat["0X3"]),
            ("OT", HierCat["OT"]),
        ]
        assert len(HierCat) == 11

    def test_df(self, HierCat: climate_categories.HierarchicalCategorization):
        expected = pd.DataFrame(
            index=["0", "1", "2"],
            data={
                "title": ["Category 0", "Category 1", "Category 2"],
                "comment": [
                    "Top-most category",
                    "The first category",
                    "The second category",
                ],
                "alternative_codes": [("TOTAL",), tuple(), tuple()],
                "children": [
                    (("1", "2", "3"), ("0X3", "3"), ("1A", "1B", "2", "3")),
                    (("1A", "1B"),),
                    (("2A", "2B"),),
                ],
            },
        )
        pd.testing.assert_frame_equal(HierCat.df.head(3), expected)

    def test_extend(self, HierCat: climate_categories.HierarchicalCategorization):
        HierCat_ext = HierCat.extend(
            name="ext",
            categories={
                "2A1": {"title": "Category 2A1"},
                "2A2": {"title": "Category 2A2"},
            },
            children=[("2A", ("2A1", "2A2")), ("2", ("2A1", "2A2", "2B"))],
        )
        assert HierCat_ext.name == "HierCat_ext"
        assert HierCat_ext.title == "Hierarchical Categorization + ext"
        assert HierCat_ext.references == ""
        assert HierCat_ext.last_update == datetime.date.today()
        assert HierCat_ext.institution == ""
        assert HierCat_ext.version == "one"
        assert (
            HierCat_ext.comment
            == "A simple hierarchical categorization with categories with relationships"
            " extended by ext"
        )
        assert HierCat_ext.hierarchical is True

        assert HierCat_ext["2"].children == [
            {HierCat_ext["2A"], HierCat_ext["2B"]},
            {HierCat_ext["2A1"], HierCat_ext["2A2"], HierCat_ext["2B"]},
        ]
        assert HierCat_ext["2A1"].level == 4
        assert HierCat_ext["2A1"].title == "Category 2A1"
        assert HierCat_ext["2A1"].comment is None
        assert HierCat_ext["2A1"].children == []

        assert "0" in HierCat_ext
        assert list(HierCat_ext.all_keys()) == [
            "0",
            "TOTAL",
            "1",
            "2",
            "3",
            "1A",
            "1a",
            "1B",
            "1b",
            "2A",
            "2a",
            "2B",
            "2b",
            "3A",
            "3a",
            "0X3",
            "0E3",
            "OT",
            "2A1",
            "2A2",
        ]
        assert list(HierCat_ext.keys()) == [
            "0",
            "1",
            "2",
            "3",
            "1A",
            "1B",
            "2A",
            "2B",
            "3A",
            "0X3",
            "OT",
            "2A1",
            "2A2",
        ]
        assert len(HierCat_ext) == 13

    def test_extend_new_alternatives(
        self, HierCat: climate_categories.HierarchicalCategorization
    ):
        ext = HierCat.extend(name="ext", alternative_codes={"yksi": "1"})
        assert ext["yksi"] == HierCat["1"]

    def test_roundtrip(self, HierCat, tmpdir):
        f = pathlib.Path(tmpdir) / "thing.yaml"
        HierCat.to_yaml(f)
        ReRead = climate_categories.HierarchicalCategorization.from_yaml(f)
        assert HierCat.name == ReRead.name
        assert HierCat.title == ReRead.title
        assert HierCat.comment == ReRead.comment
        assert HierCat.references == ReRead.references
        assert HierCat.hierarchical == ReRead.hierarchical
        assert HierCat.total_sum == ReRead.total_sum
        assert HierCat.last_update == ReRead.last_update
        assert HierCat.version == ReRead.version
        for key in HierCat:
            assert key in ReRead
            assert HierCat[key] == ReRead[key]
        assert HierCat == ReRead

    def test_level_error(self, HierCat: climate_categories.HierarchicalCategorization):
        HierCat.canonical_top_level_category = None
        with pytest.raises(ValueError, match="Can not calculate the level"):
            HierCat["1"].level

    def test_parents_code(self, HierCat):
        assert HierCat.parents(HierCat["1"]) == HierCat.parents("1")

    def test_children_code(self, HierCat):
        assert HierCat.children(HierCat["1"]) == HierCat.children("1")


def test_compare(HierCat, SimpleCat):
    assert HierCat != SimpleCat
    assert HierCat != "HierCat"
    ext = HierCat.extend(name="ext")
    assert HierCat != ext


class TestShowAsTree:
    def test_show_as_tree(self, HierCat):
        assert (
            HierCat.show_as_tree()
            == """0 Category 0
╠╤══ ('0 Category 0's children, option 1)
║├1 Category 1
║│├1A Category 1A
║│╰1B Category 1B
║├2 Category 2
║│├2A Category 2A
║│╰2B Category 2B
║╰3 Category 3
║ ╰3A Category 3A
╠╕ ('0 Category 0's children, option 2)
║├0X3 Total excluding category 3
║│├1 Category 1
║││├1A Category 1A
║││╰1B Category 1B
║│╰2 Category 2
║│ ├2A Category 2A
║│ ╰2B Category 2B
║╰3 Category 3
║ ╰3A Category 3A
╠╕ ('0 Category 0's children, option 3)
║├1A Category 1A
║├1B Category 1B
║├2 Category 2
║│├2A Category 2A
║│╰2B Category 2B
║╰3 Category 3
║ ╰3A Category 3A
╚═══

OT Other top category
├1B Category 1B
╰2B Category 2B
"""
        )

    def test_format_func(self, HierCat):
        assert (
            HierCat.show_as_tree(format_func=lambda x: x.codes[0])
            == """0
╠╤══ ('0's children, option 1)
║├1
║│├1A
║│╰1B
║├2
║│├2A
║│╰2B
║╰3
║ ╰3A
╠╕ ('0's children, option 2)
║├0X3
║│├1
║││├1A
║││╰1B
║│╰2
║│ ├2A
║│ ╰2B
║╰3
║ ╰3A
╠╕ ('0's children, option 3)
║├1A
║├1B
║├2
║│├2A
║│╰2B
║╰3
║ ╰3A
╚═══

OT
├1B
╰2B
"""
        )

    def test_maxdepth(self, HierCat):
        assert (
            HierCat.show_as_tree(maxdepth=2)
            == """0 Category 0
╠╤══ ('0 Category 0's children, option 1)
║├1 Category 1
║├2 Category 2
║╰3 Category 3
╠╕ ('0 Category 0's children, option 2)
║├0X3 Total excluding category 3
║╰3 Category 3
╠╕ ('0 Category 0's children, option 3)
║├1A Category 1A
║├1B Category 1B
║├2 Category 2
║╰3 Category 3
╚═══

OT Other top category
├1B Category 1B
╰2B Category 2B
"""
        )

    def test_root(self, HierCat):
        assert (
            HierCat.show_as_tree(root="1")
            == """1 Category 1
├1A Category 1A
╰1B Category 1B
"""
        )


class TestIO:
    def test_spec_misses_hierarchical(self, spec_simple):
        with pytest.raises(KeyError):
            del spec_simple["hierarchical"]
            climate_categories.Categorization.from_spec(spec_simple)

    def test_spec_wrong_hierarchical(self, spec_simple, spec_hier):
        with pytest.raises(
            ValueError, match="Specification is for a hierarchical categorization"
        ):
            climate_categories.Categorization.from_spec(spec_hier)

        with pytest.raises(
            ValueError, match="Specification is for a non-hierarchical categorization"
        ):
            climate_categories.HierarchicalCategorization.from_spec(spec_simple)

    def test_from_spec(self, spec_simple, SimpleCat):
        fs = climate_categories.from_spec(spec_simple)
        assert fs == SimpleCat
        assert fs.keys() == SimpleCat.keys()
        assert list(fs.values()) == list(SimpleCat.values())

    def test_roundtrip(self, tmpdir, any_cat):
        any_cat.to_yaml(tmpdir / "any_cat.yaml")
        any_cat_r = climate_categories.from_yaml(tmpdir / "any_cat.yaml")
        assert any_cat == any_cat_r
        assert list(any_cat.values()) == list(any_cat_r.values())

        any_cat.to_pickle(tmpdir / "any_cat.pickle")
        any_cat_p = climate_categories.from_pickle(tmpdir / "any_cat.pickle")
        assert any_cat == any_cat_p
        assert list(any_cat.values()) == list(any_cat_p.values())

        assert climate_categories.from_pickle(
            tmpdir / "any_cat.pickle"
        ) == climate_categories.Categorization.from_pickle(tmpdir / "any_cat.pickle")

    def test_roundtrip_hierarchical(self, tmpdir, HierCat):
        HierCat.to_yaml(tmpdir / "HierCat.yaml")
        HierCat_r = climate_categories.HierarchicalCategorization.from_yaml(
            tmpdir / "HierCat.yaml"
        )
        assert HierCat == HierCat_r

    def test_broken(self):
        with pytest.raises(
            strictyaml.YAMLValidationError,
            match="unexpected key not in schema 'reefrences'",
        ):
            climate_categories.HierarchicalCategorization.from_yaml(
                importlib.resources.open_text(
                    climate_categories.tests.data,
                    "broken_hierarchical_categorization.yaml",
                )
            )

    def test_broken_hierarchical(self):
        with pytest.raises(ValueError, match="'hierarchical' must be "):
            climate_categories.from_yaml(
                importlib.resources.open_text(
                    climate_categories.tests.data,
                    "broken_simple_categorization.yaml",
                )
            )



import pathlib
import shutil

import requests


def download_cached(url: str, fpath: pathlib.Path):
    if not fpath.exists():
        print(f"{fpath} not found, downloading it.")
        r = requests.get(url, stream=True)
        if r.status_code == 200:
            with fpath.open("wb") as f:
                r.raw.decode_content = True
                shutil.copyfileobj(r.raw, f)
        else:
            raise FileNotFoundError(f"Could not download {url!r}")


def title_case(instr: str) -> str:
    return (
        instr.title()
        .replace(" And ", " and ")
        .replace(" Of ", " of ")
        .replace(" On ", " on ")
        .replace(" Or ", " or ")
        .replace(" To ", " to ")
        .replace(" As ", " as ")
        .replace(" For ", " for ")
        .replace(" From ", " from ")
        .replace(" With ", " with ")
        .replace(" Without ", " without ")
        .replace("Nox", "NOx")
        .replace("Nh3", "NH3")
        .replace("Co2", "CO2")
        .replace("Sf6", "SF6")
        .replace("Pfc", "PFC")
        .replace("Tft", "TFT")
    )



"""Run this via `make climate_categories/data/RCMIP.yaml` in the main directory."""

import pathlib

import pandas as pd
from utils import download_cached

import climate_categories

RCMIP_SUBMISSIONS_TEMPLATE = "rcmip-data-submission-template.xlsx"
URL = (
    "https://gitlab.com/rcmip/rcmip/-/raw/master/data/submission-template/"
    + RCMIP_SUBMISSIONS_TEMPLATE
)
INPATH = pathlib.Path(f"./data_generation/{RCMIP_SUBMISSIONS_TEMPLATE}")
OUTPATH = pathlib.Path("./climate_categories/data/RCMIP.yaml")

comment = """AFOLU in the SSPDb is AFOLU minus any agriculture
related fossil fuel based emissions hence is not the same as the
WG3 AFOLU definition. Rather AFOLU in the SSPDb is AFOLU as expected by
MAGICC (i.e. exluding agriculture related fossil fuel use), hence
we call it MAGICC AFOLU."""

titles = {
    "BC": "Black carbon emissions",
    "BC|MAGICC AFOLU": "Black carbon AFOLU emissions",
    "BC|MAGICC Fossil and Industrial": "Black carbon fossil and industrial emissions",
    "BC|Other": "Black carbon emissions from other sources",
    "CH4": "Methane emissions",
    "CH4|MAGICC AFOLU": "Methane AFOLU emissions",
    "CH4|MAGICC Fossil and Industrial": "Methane fossil and industrial emissions",
    "CH4|Other": "Methane emissions from other sources",
    "CO": "Carbon monoxide emissions",
    "CO|MAGICC AFOLU": "Carbon monoxide AFOLU emissions",
    "CO|MAGICC Fossil and Industrial": (
        "Carbon monoxide fossil " "and industrial emissions"
    ),
    "CO|Other": "Carbon monoxide emissions from other sources",
    "CO2": "Carbon dioxide emissions",
    "CO2|MAGICC AFOLU": "Carbon dioxide AFOLU emissions",
    "CO2|MAGICC Fossil and Industrial": (
        "Carbon dioxide fossil " "and industrial emissions"
    ),
    "CO2|Other": "Carbon dioxide emissions from other sources",
    "F-Gases": "F-gas emissions",
    "F-Gases|HFC": "Hydrofluorocarbons (HFCs and HCFCs) emissions",
    "F-Gases|HFC|HFC125": "HFC125 emissions",
    "F-Gases|HFC|HFC134a": "HFC134a emissions",
    "F-Gases|HFC|HFC143a": "HFC143a emissions",
    "F-Gases|HFC|HFC152a": "HFC152a emissions",
    "F-Gases|HFC|HFC227ea": "HFC227ea emissions",
    "F-Gases|HFC|HFC23": "HFC23 emissions",
    "F-Gases|HFC|HFC236fa": "HFC236fa emissions",
    "F-Gases|HFC|HFC245fa": "HFC245fa emissions",
    "F-Gases|HFC|HFC32": "HFC32 emissions",
    "F-Gases|HFC|HFC365mfc": "HFC365mfc emissions",
    "F-Gases|HFC|HFC4310mee": "HFC43-10mee emissions",
    "F-Gases|NF3": "Nitrogen trifluoride emissions",
    "F-Gases|PFC": "Perfluorocarbons  emissions",
    "F-Gases|PFC|C2F6": "C2F6 emissions",
    "F-Gases|PFC|C3F8": "C3F8 emissions",
    "F-Gases|PFC|C4F10": "C4F10 emissions",
    "F-Gases|PFC|C5F12": "C5F12 emissions",
    "F-Gases|PFC|C6F14": "C6F14 emissions",
    "F-Gases|PFC|C7F16": "C7F16 emissions",
    "F-Gases|PFC|C8F18": "C8F18 emissions",
    "F-Gases|PFC|cC4F8": "c-C4F8 emissions",
    "F-Gases|PFC|CF4": "CF4 emissions",
    "F-Gases|SF6": "Sulfur hexafluoride emissions",
    "F-Gases|SO2F2": "Sulfuryl fluoride emissions",
    "Montreal Gases": "Montreal gas emissions",
    "Montreal Gases|CCl4": "CCl4 emissions",
    "Montreal Gases|CFC": "CFC emissions",
    "Montreal Gases|CFC|CFC11": "CFC11 emissions",
    "Montreal Gases|CFC|CFC113": "CFC113 emissions",
    "Montreal Gases|CFC|CFC114": "CFC114 emissions",
    "Montreal Gases|CFC|CFC115": "CFC115 emissions",
    "Montreal Gases|CFC|CFC12": "CFC12 emissions",
    "Montreal Gases|CH2Cl2": "CH2Cl2 emissions",
    "Montreal Gases|CH3Br": "CH3Br emissions",
    "Montreal Gases|CH3CCl3": "CH3CCl3 emissions",
    "Montreal Gases|CH3Cl": "CH3Cl emissions",
    "Montreal Gases|CHCl3": "CHCl3 emissions",
    "Montreal Gases|Halon1202": "Halon-1202 emissions",
    "Montreal Gases|Halon1211": "Halon-1211 emissions",
    "Montreal Gases|Halon1301": "Halon-1301 emissions",
    "Montreal Gases|Halon2402": "Halon-2402 emissions",
    "Montreal Gases|HCFC141b": "HCFC141b emissions",
    "Montreal Gases|HCFC142b": "HCFC22 emissions",
    "Montreal Gases|HCFC22": "HCFC22 emissions",
    "N2O": "Nitrogen emissions",
    "N2O|MAGICC AFOLU": "Nitrogen AFOLU emissions",
    "N2O|MAGICC Fossil and Industrial": "Nitrogen fossil and industrial emissions",
    "N2O|Other": "Nitrogen emissions from other sources",
    "NH3": "Ammonia emissions",
    "NH3|MAGICC AFOLU": "Ammonia AFOLU emissions",
    "NH3|MAGICC Fossil and Industrial": "Ammonia fossil and industrial emissions",
    "NH3|Other": "Ammonia emissions from other sources",
    "NOx": "Nitrous oxide emissions",
    "NOx|MAGICC AFOLU": "Nitrous oxide AFOLU emissions",
    "NOx|MAGICC Fossil and Industrial": "Nitrous oxide fossil and industrial emissions",
    "NOx|Other": "Nitrous oxide emissions from other sources",
    "OC": "Organic carbon emissions",
    "OC|MAGICC AFOLU": "Organic carbon AFOLU emissions",
    "OC|MAGICC Fossil and Industrial": "Organic carbon fossil and industrial emissions",
    "OC|Other": "Organic carbon emissions from other sources",
    "Sulfur": "Sulfur emissions",
    "Sulfur|MAGICC AFOLU": "Sulfur AFOLU emissions",
    "Sulfur|MAGICC Fossil and Industrial": "Sulfur fossil and industrial emissions",
    "Sulfur|Other": "Sulfur emissions from other sources",
    "VOC": "(Non-methane) volatile organic compounds emissions",
    "VOC|MAGICC AFOLU": "(Non-methane) volatile organic compounds AFOLU emissions",
    "VOC|MAGICC Fossil and Industrial": (
        "(Non-methane) volatile organic compounds" " fossil and industrial emissions"
    ),
    "VOC|Other": (
        "(Non-methane) volatile organic compounds" " emissions from other sources"
    ),
}


def main():
    download_cached(URL, INPATH)

    categories = {}

    categories["Emissions"] = {"title": "RCMIP Emissions", "children": [[]]}

    definitions = pd.read_excel(INPATH, sheet_name="variable_definitions")

    for _, item in definitions[definitions.Category == "Emissions"].iterrows():

        species = item.Variable
        print(species)
        parent = item.Variable.rsplit("|", maxsplit=1)[0]
        print(parent)
        item_comment = item.Definition.replace(
            "(please provide a definition of other sources in this category in the "
            "'comments' tab)",
            "(please provide a definition)",
        )
        if species not in categories:
            categories[species] = {
                "title": titles[species.rsplit("Emissions|", maxsplit=1)[1]],
                "comment": item_comment,
                "children": [[]],
            }
        if species not in categories[parent]["children"][0]:
            categories[parent]["children"][0].append(species)

    spec = {
        "name": "RCMIP",
        "title": (
            "Emissions categories from the Reduced Complexity Model "
            "Intercomparison Project (RCMIP)"
        ),
        "comment": comment,
        "references": (
            "Nicholls, Z. R. J., Meinshausen, M., Lewis, J., Gieseke, R., "
            "Dommenget, D., Dorheim, K., Fan, C.-S., Fuglestvedt, J. S., Gasser, T., "
            "Golüke, U., Goodwin, P., Hartin, C., Hope, A. P., Kriegler, E., Leach, "
            "N. J., Marchegiani, D., McBride, L. A., Quilcaille, Y., Rogelj, J., "
            "Salawitch, R. J., Samset, B. H., Sandstad, M., Shiklomanov, A. N., "
            "Skeie, R. B., Smith, C. J., Smith, S., Tanaka, K., Tsutsui, J., "
            "and Xie, Z.: Reduced Complexity Model Intercomparison Project Phase 1:"
            " introduction and evaluation of global-mean temperature response, "
            "Geosci. Model Dev., 13, 5175–5190, "
            "https://doi.org/10.5194/gmd-13-5175-2020, 2020."
        ),
        "institution": "RCMIP",
        "last_update": "2020-09-21",
        "hierarchical": True,
        "version": "v5.1.0",
        "total_sum": True,
        "categories": categories,
        "canonical_top_level_category": "Emissions",
    }

    RCMIP = climate_categories.HierarchicalCategorization.from_spec(spec)

    RCMIP.to_yaml(OUTPATH)

    climate_categories.HierarchicalCategorization.from_yaml(OUTPATH)


if __name__ == "__main__":
    main()



import climate_categories


def test_number_of_categories():
    assert len(climate_categories.IPCC1996) == 233
    assert len(climate_categories.IPCC2006) == 290


def test_levels():
    assert climate_categories.IPCC1996.level("1") == 2
    assert climate_categories.IPCC1996.level("1A") == 3
    assert climate_categories.IPCC2006.level("1") == 2
    assert climate_categories.IPCC2006.level("1A") == 3


def test_consistent():
    for cat in climate_categories.IPCC2006.values():
        if "corresponding_categories_IPCC1996" in cat.info:
            for ccat in cat.info["corresponding_categories_IPCC1996"]:
                assert ccat in climate_categories.IPCC1996


def test_conversion():
    conv = climate_categories.IPCC1996.conversion_to(climate_categories.IPCC2006)

    problems = conv.find_over_counting_problems()
    problematic_categories = list(
        sorted([problem.category.codes[0] for problem in problems])
    )
    assert problematic_categories == [
        "2.B.7",
        "2.B.9",
        "2.B.9.a",
        "2.B.9.b",
        "6.A.1",
        "6.A.2",
        "6.A.3",
        "6.B.3",
    ]

    # still a lot of unmapped categories )-:
    missing96, missing06 = conv.find_unmapped_categories()
    assert len(missing96) == 71
    assert len(missing06) == 110



import climate_categories


def test_levels_crf():
    assert climate_categories.CRFDI.level("1") == 2
    assert climate_categories.CRFDI.level("1.A") == 3


def test_lulucf_crf():
    assert climate_categories.CRFDI["4"] in climate_categories.CRFDI["8677"].children[0]
    assert (
        climate_categories.CRFDI["4"]
        not in climate_categories.CRFDI["10464"].children[0]
    )


def test_levels_bur():
    assert climate_categories.BURDI.level("1") == 2
    assert climate_categories.BURDI.level("1.A") == 3


def test_lulucf_bur():
    assert (
        climate_categories.BURDI["5"] in climate_categories.BURDI["24540"].children[0]
    )
    assert (
        climate_categories.BURDI["5"]
        not in climate_categories.BURDI["15163"].children[0]
    )
